@@ -3,43 +3,32 @@ from json import dumps as json_dumps
 from json import loads as json_loads
 from os import makedirs
 from os.path import basename
+from os.path import dirname
 from os.path import isdir
 from os.path import join as path_join
 from shutil import copy
 from shutil import move
 from shutil import rmtree
+from sqlite3 import Connection
 from sqlite3 import DatabaseError
 from sqlite3 import OperationalError
+from sqlite3 import connect as connect_database
+from typing import Collection
 from typing import List
 from typing import Optional
 from typing import Tuple
+from typing import Union
 
 from .__version__ import __version__
-from .database import Connection
-from .database import connect_database
-from .database import count
-from .database import insert
-from .database import make_tables
-from .database import select
-from .database import select_all
-from .database import tiered_path
-from .database import update
-from .journals import journals_table
-from .settings import add_history
-from .settings import read_setting
-from .settings import settings_table
-from .settings import write_setting
-from .submissions import submissions_table
-from .users import users_table
 
 
 def get_version(db: Connection) -> str:
     try:
         # Database version 3.0.0 and above
-        return read_setting(db, "VERSION")
+        return db.execute(f"SELECT SVALUE FROM SETTINGS WHERE SETTING = 'VERSION'").fetchone()[0]
     except OperationalError:
         # Database version 2.7.0
-        return next(select(db, "INFOS", ["VALUE"], ["FIELD"], ["VERSION"]))[0]
+        return db.execute(f"SELECT VALUE FROM SETTINGS WHERE FIELD = 'VERSION'").fetchone()[0]
 
 
 def compare_versions(a: str, b: str) -> int:
@@ -56,6 +45,44 @@ def compare_versions(a: str, b: str) -> int:
     return 0
 
 
+def insert(db: Connection, table: str, keys: Collection[str], values: Collection[Union[int, str]],
+           replace: bool = True):
+    db.execute(
+        f"""INSERT OR {"REPLACE" if replace else "IGNORE"} INTO {table}
+        ({",".join(keys)})
+        VALUES ({",".join(["?"] * len(values))})""",
+        values
+    )
+
+
+def update(db: Connection, table: str, fields: Collection[str], values: Collection[Union[int, str]], key: str,
+           key_value: str):
+    assert len(fields) == len(values) and len(fields) > 0
+
+    update_values: List[str] = [f"{u} = ?" for u in fields]
+
+    db.execute(
+        f"""UPDATE {table}
+        SET {",".join(update_values)}
+        WHERE {key} = ?""",
+        (*values, key_value,)
+    )
+
+
+def count(db: Connection, table: str) -> int:
+    return db.execute(f"SELECT COUNT(*) FROM {table}").fetchall()[0][0]
+
+
+def database_path(db: Connection) -> Optional[str]:
+    name: str
+    filename: Optional[str]
+    for _, name, filename in db.execute("PRAGMA database_list"):
+        if name == "main" and filename is not None:
+            return filename
+
+    return None
+
+
 def make_database_3(db: Connection):
     # Create submissions table
     db.execute(
@@ -153,18 +180,19 @@ def make_database_3_1(db: Connection):
     db.commit()
 
     # Add settings
-    insert(db, "SETTINGS", ["SETTING", "SVALUE"], ["USRN", "0"], False)
-    insert(db, "SETTINGS", ["SETTING", "SVALUE"], ["SUBN", "0"], False)
-    insert(db, "SETTINGS", ["SETTING", "SVALUE"], ["LASTUPDATE", "0"], False)
-    insert(db, "SETTINGS", ["SETTING", "SVALUE"], ["LASTSTART", "0"], False)
-    insert(db, "SETTINGS", ["SETTING", "SVALUE"], ["COOKIES", "{}"], False)
-    insert(db, "SETTINGS", ["SETTING", "SVALUE"], ["FILESFOLDER", "FA.files"], False)
-    insert(db, "SETTINGS", ["SETTING", "SVALUE"], ["VERSION", "3.1.0"], False)
+    db.execute(f"INSERT OR IGNORE INTO SETTINGS (SETTING, SVALUE) VALUES (?, ?)", ["USRN", "0"])
+    db.execute(f"INSERT OR IGNORE INTO SETTINGS (SETTING, SVALUE) VALUES (?, ?)", ["SUBN", "0"])
+    db.execute(f"INSERT OR IGNORE INTO SETTINGS (SETTING, SVALUE) VALUES (?, ?)", ["JRNN", "0"])
+    db.execute(f"INSERT OR IGNORE INTO SETTINGS (SETTING, SVALUE) VALUES (?, ?)", ["LASTUPDATE", "0"])
+    db.execute(f"INSERT OR IGNORE INTO SETTINGS (SETTING, SVALUE) VALUES (?, ?)", ["LASTSTART", "0"])
+    db.execute(f"INSERT OR IGNORE INTO SETTINGS (SETTING, SVALUE) VALUES (?, ?)", ["COOKIES", "{}"])
+    db.execute(f"INSERT OR IGNORE INTO SETTINGS (SETTING, SVALUE) VALUES (?, ?)", ["FILESFOLDER", "FA.files"])
+    db.execute(f"INSERT OR IGNORE INTO SETTINGS (SETTING, SVALUE) VALUES (?, ?)", ["VERSION", "3.1.0"])
 
     db.commit()
 
 
-def make_database_3_2(db: Connection) -> Connection:
+def make_database_3_2(db: Connection):
     db.execute(
         f"""CREATE TABLE IF NOT EXISTS USERS
         (USERNAME TEXT UNIQUE NOT NULL,
@@ -212,29 +240,153 @@ def make_database_3_2(db: Connection) -> Connection:
         PRIMARY KEY (SETTING ASC));"""
     )
 
-    insert(db, settings_table, ["SETTING", "SVALUE"], ["USRN", "0"], False)
-    insert(db, settings_table, ["SETTING", "SVALUE"], ["SUBN", "0"], False)
-    insert(db, settings_table, ["SETTING", "SVALUE"], ["JRNN", "0"], False)
-    insert(db, settings_table, ["SETTING", "SVALUE"], ["LASTUPDATE", "0"], False)
-    insert(db, settings_table, ["SETTING", "SVALUE"], ["LASTSTART", "0"], False)
-    insert(db, settings_table, ["SETTING", "SVALUE"], ["COOKIES", "{}"], False)
-    insert(db, settings_table, ["SETTING", "SVALUE"], ["FILESFOLDER", "FA.files"], False)
-    insert(db, settings_table, ["SETTING", "SVALUE"], ["VERSION", __version__], False)
+    db.execute(f"INSERT OR IGNORE INTO SETTINGS (SETTING, SVALUE) VALUES (?, ?)", ["USRN", "0"])
+    db.execute(f"INSERT OR IGNORE INTO SETTINGS (SETTING, SVALUE) VALUES (?, ?)", ["SUBN", "0"])
+    db.execute(f"INSERT OR IGNORE INTO SETTINGS (SETTING, SVALUE) VALUES (?, ?)", ["JRNN", "0"])
+    db.execute(f"INSERT OR IGNORE INTO SETTINGS (SETTING, SVALUE) VALUES (?, ?)", ["LASTUPDATE", "0"])
+    db.execute(f"INSERT OR IGNORE INTO SETTINGS (SETTING, SVALUE) VALUES (?, ?)", ["LASTSTART", "0"])
+    db.execute(f"INSERT OR IGNORE INTO SETTINGS (SETTING, SVALUE) VALUES (?, ?)", ["COOKIES", "{}"])
+    db.execute(f"INSERT OR IGNORE INTO SETTINGS (SETTING, SVALUE) VALUES (?, ?)", ["FILESFOLDER", "FA.files"])
+    db.execute(f"INSERT OR IGNORE INTO SETTINGS (SETTING, SVALUE) VALUES (?, ?)", ["VERSION", "3.2.0"])
 
-    return db
+    db.commit()
+
+
+def make_database_3_3(db: Connection):
+    db.execute(
+        f"""CREATE TABLE IF NOT EXISTS USERS
+        (USERNAME TEXT UNIQUE NOT NULL,
+        FOLDERS TEXT NOT NULL,
+        GALLERY TEXT,
+        SCRAPS TEXT,
+        FAVORITES TEXT,
+        MENTIONS TEXT,
+        JOURNALS TEXT,
+        PRIMARY KEY (USERNAME ASC));"""
+    )
+
+    db.execute(
+        f"""CREATE TABLE IF NOT EXISTS SUBMISSIONS
+        (ID INT UNIQUE NOT NULL,
+        AUTHOR TEXT NOT NULL,
+        TITLE TEXT,
+        UDATE DATE NOT NULL,
+        DESCRIPTION TEXT,
+        TAGS TEXT,
+        CATEGORY TEXT,
+        SPECIES TEXT,
+        GENDER TEXT,
+        RATING TEXT,
+        FILELINK TEXT,
+        FILEEXT TEXT,
+        FILESAVED INT,
+        PRIMARY KEY (ID ASC));"""
+    )
+
+    db.execute(
+        f"""CREATE TABLE IF NOT EXISTS JOURNALS
+        (ID INT UNIQUE NOT NULL,
+        AUTHOR TEXT NOT NULL,
+        TITLE TEXT,
+        UDATE DATE NOT NULL,
+        CONTENT TEXT,
+        PRIMARY KEY (ID ASC));"""
+    )
+
+    db.execute(
+        f"""CREATE TABLE IF NOT EXISTS SETTINGS
+        (SETTING TEXT UNIQUE,
+        SVALUE TEXT,
+        PRIMARY KEY (SETTING ASC));"""
+    )
+
+    db.execute(f"INSERT OR IGNORE INTO SETTINGS (SETTING, SVALUE) VALUES (?, ?)", ["USRN", "0"])
+    db.execute(f"INSERT OR IGNORE INTO SETTINGS (SETTING, SVALUE) VALUES (?, ?)", ["SUBN", "0"])
+    db.execute(f"INSERT OR IGNORE INTO SETTINGS (SETTING, SVALUE) VALUES (?, ?)", ["JRNN", "0"])
+    db.execute(f"INSERT OR IGNORE INTO SETTINGS (SETTING, SVALUE) VALUES (?, ?)", ["HISTORY", "[]"])
+    db.execute(f"INSERT OR IGNORE INTO SETTINGS (SETTING, SVALUE) VALUES (?, ?)", ["COOKIES", "{}"])
+    db.execute(f"INSERT OR IGNORE INTO SETTINGS (SETTING, SVALUE) VALUES (?, ?)", ["FILESFOLDER", "FA.files"])
+    db.execute(f"INSERT OR IGNORE INTO SETTINGS (SETTING, SVALUE) VALUES (?, ?)", ["VERSION", "3.3.0"])
+
+    db.commit()
+
+
+def make_database_4(db: Connection):
+    db.execute(
+        f"""CREATE TABLE IF NOT EXISTS USERS
+        (USERNAME TEXT UNIQUE NOT NULL CHECK (length(USERNAME) > 0),
+        FOLDERS TEXT NOT NULL,
+        GALLERY TEXT NOT NULL,
+        SCRAPS TEXT NOT NULL,
+        FAVORITES TEXT NOT NULL,
+        MENTIONS TEXT NOT NULL,
+        JOURNALS TEXT NOT NULL,
+        PRIMARY KEY (USERNAME ASC));"""
+    )
+
+    db.execute(
+        f"""CREATE TABLE IF NOT EXISTS SUBMISSIONS
+        (ID INT UNIQUE NOT NULL CHECK (ID > 0),
+        AUTHOR TEXT NOT NULL CHECK (length(AUTHOR) > 0),
+        TITLE TEXT NOT NULL,
+        DATE DATE NOT NULL CHECK (DATE==strftime('%Y-%m-%d',DATE)),
+        DESCRIPTION TEXT NOT NULL,
+        TAGS TEXT NOT NULL,
+        CATEGORY TEXT NOT NULL,
+        SPECIES TEXT NOT NULL,
+        GENDER TEXT NOT NULL,
+        RATING TEXT NOT NULL,
+        FILELINK TEXT NOT NULL,
+        FILEEXT TEXT NOT NULL,
+        FILESAVED INT NOT NULL CHECK (FILESAVED in (0, 1)),
+        PRIMARY KEY (ID ASC));"""
+    )
+
+    db.execute(
+        f"""CREATE TABLE IF NOT EXISTS JOURNALS
+        (ID INT UNIQUE NOT NULL CHECK (ID > 0),
+        AUTHOR TEXT NOT NULL CHECK (length(AUTHOR) > 0),
+        TITLE TEXT NOT NULL,
+        DATE DATE NOT NULL CHECK (DATE==strftime('%Y-%m-%d',DATE)),
+        CONTENT TEXT NOT NULL,
+        PRIMARY KEY (ID ASC));"""
+    )
+
+    db.execute(
+        f"""CREATE TABLE IF NOT EXISTS SETTINGS
+        (SETTING TEXT UNIQUE NOT NULL CHECK (length(SETTING) > 0),
+        SVALUE TEXT NOT NULL CHECK (length(SVALUE) > 0),
+        PRIMARY KEY (SETTING ASC));"""
+    )
+
+    # Add settings
+    db.execute(f"INSERT OR IGNORE INTO SETTINGS (SETTING, SVALUE) VALUES (?, ?)", ["USRN", "0"])
+    db.execute(f"INSERT OR IGNORE INTO SETTINGS (SETTING, SVALUE) VALUES (?, ?)", ["SUBN", "0"])
+    db.execute(f"INSERT OR IGNORE INTO SETTINGS (SETTING, SVALUE) VALUES (?, ?)", ["JRNN", "0"])
+    db.execute(f"INSERT OR IGNORE INTO SETTINGS (SETTING, SVALUE) VALUES (?, ?)", ["HISTORY", "[]"])
+    db.execute(f"INSERT OR IGNORE INTO SETTINGS (SETTING, SVALUE) VALUES (?, ?)", ["COOKIES", "{}"])
+    db.execute(f"INSERT OR IGNORE INTO SETTINGS (SETTING, SVALUE) VALUES (?, ?)", ["FILESFOLDER", "FA.files"])
+    db.execute(f"INSERT OR IGNORE INTO SETTINGS (SETTING, SVALUE) VALUES (?, ?)", ["VERSION", "4.0.0"])
+
+    db.commit()
 
 
 def update_2_7_to_3(db: Connection) -> Connection:
+    from .database import tiered_path
+
     print("Updating 2.7.0 to 3.0.0")
     db_new: Optional[Connection] = None
 
+    db_path: str = dp if (dp := database_path(db)) else "FA.db"
+    db_new_path: str = path_join(dirname(db_path), "new_" + basename(db_path))
+
     try:
-        db_new = connect_database("FA_new.db")
+        db_new = connect_database(db_new_path)
         make_database_3(db_new)
 
         # Transfer common submissions and users data
         print("Transfer common submissions and users data")
-        db.execute("ATTACH DATABASE 'FA_new.db' AS db_new")
+        db.execute(f"ATTACH DATABASE '{db_new_path}' AS db_new")
         db.execute(
             """INSERT OR IGNORE INTO
             db_new.SUBMISSIONS(ID, AUTHOR, UDATE, TITLE, DESCRIPTION, TAGS, CATEGORY, SPECIES, GENDER, RATING, FILELINK)
@@ -256,7 +408,7 @@ def update_2_7_to_3(db: Connection) -> Connection:
         user: str
         folders: str
         user_n: int = 0
-        for user, folders in select_all(db_new, "USERS", ["USERNAME", "FOLDERS"]):
+        for user, folders in db_new.execute("SELECT USERNAME, FOLDERS FROM USERS"):
             user_n += 1
             print(user_n, end="\r", flush=True)
             folders_new: List[str] = []
@@ -283,14 +435,16 @@ def update_2_7_to_3(db: Connection) -> Connection:
         print("Update submissions FILEEXT and FILESAVED and move to new location")
         sub_n: int = 0
         sub_not_found: List[int] = []
-        for id_, location in select_all(db, "SUBMISSIONS", ["ID", "LOCATION"]):
+        files_folder_old: str = path_join(dirname(db_path), "FA.files")
+        files_folder_new: str = path_join(dirname(db_path), "FA.files_new")
+        for id_, location in db.execute("SELECT ID, LOCATION FROM SUBMISSIONS"):
             sub_n += 1
             print(sub_n, end="\r", flush=True)
-            if isdir(folder_old := path_join("FA.files", location.strip("/"))):
+            if isdir(folder_old := path_join(files_folder_old, location.strip("/"))):
                 fileglob = glob(path_join(folder_old, "submission*"))
                 fileext = ""
                 if filename := fileglob[0] if fileglob else "":
-                    makedirs((folder_new := path_join("FA.files_new", tiered_path(id_))), exist_ok=True)
+                    makedirs((folder_new := path_join(files_folder_new, tiered_path(id_))), exist_ok=True)
                     copy(filename, path_join(folder_new, (filename := basename(filename))))
                     fileext = filename.split(".")[-1] if "." in filename else ""
                 else:
@@ -316,18 +470,18 @@ def update_2_7_to_3(db: Connection) -> Connection:
 
         # Replace older files folder with new
         print("Replace older files folder with new")
-        if isdir("FA.files"):
+        if isdir(f := path_join(dirname(db_path), "FA.files")):
             if not sub_not_found:
-                rmtree("FA.files")
+                rmtree(f)
             else:
                 print("Saving older FA.files to FA.files_old")
-                move("FA.files", "FA.files_old")
-        if isdir("FA.files_new"):
-            move("FA.files_new", "FA.files")
+                move(f, f + "_old")
+        if isdir(f := path_join(dirname(db_path), "FA.files_new")):
+            move(f, path_join(dirname(db_path), "FA.files"))
 
         # Update counters for new database
-        write_setting(db_new, "SUBN", str(count(db_new, "SUBMISSIONS")))
-        write_setting(db_new, "USRN", str(count(db_new, "USERS")))
+        update(db, "SETTINGS", ["SVALUE"], [str(count(db_new, "SUBMISSIONS"))], "SETTING", "SUBN")
+        update(db, "SETTINGS", ["SVALUE"], [str(count(db_new, "USERS"))], "SETTING", "USRN")
 
         # Close databases and replace old database
         print("Close databases and replace old database")
@@ -335,8 +489,8 @@ def update_2_7_to_3(db: Connection) -> Connection:
         db.close()
         db_new.commit()
         db_new.close()
-        move("FA.db", "FA_2_7.db")
-        move("FA_new.db", "FA.db")
+        move(db_path, path_join(dirname(db_path), "v2_7_" + basename(db_path)))
+        move(db_new_path, db_path)
     except (BaseException, Exception) as err:
         print("Database update interrupted!")
         db.commit()
@@ -346,20 +500,23 @@ def update_2_7_to_3(db: Connection) -> Connection:
             db_new.close()
         raise err
 
-    return connect_database("FA.db")
+    return connect_database(db_path)
 
 
 def update_3_to_3_1(db: Connection) -> Connection:
     print("Updating 3.0.0 to 3.1.0")
     db_new: Optional[Connection] = None
 
+    db_path: str = dp if (dp := database_path(db)) else "FA.db"
+    db_new_path: str = path_join(dirname(db_path), "new_" + basename(db_path))
+
     try:
-        db_new = connect_database("FA_new.db")
+        db_new = connect_database(db_new_path)
         make_database_3_1(db_new)
 
         # Transfer common submissions and users data
         print("Transfer common submissions and users data")
-        db.execute("ATTACH DATABASE 'FA_new.db' AS db_new")
+        db.execute(f"ATTACH DATABASE '{db_new_path}' AS db_new")
         db.execute(
             """INSERT OR IGNORE INTO db_new.SUBMISSIONS
             SELECT * FROM SUBMISSIONS"""
@@ -370,9 +527,8 @@ def update_3_to_3_1(db: Connection) -> Connection:
         )
         db.execute(
             """INSERT OR REPLACE INTO db_new.SETTINGS
-            SELECT * FROM SETTINGS"""
+            SELECT * FROM SETTINGS WHERE SETTING != "VERSION";"""
         )
-        db.execute("UPDATE db_new.SETTINGS SET SVALUE = '3.1.0' WHERE SETTING = 'VERSION'")
 
         db.commit()
         db_new.commit()
@@ -388,8 +544,8 @@ def update_3_to_3_1(db: Connection) -> Connection:
         db.close()
         db_new.commit()
         db_new.close()
-        move("FA.db", "FA_3.db")
-        move("FA_new.db", "FA.db")
+        move(db_path, path_join(dirname(db_path), "v3_" + basename(db_path)))
+        move(db_new_path, db_path)
     except (BaseException, Exception) as err:
         print("Database update interrupted!")
         db.commit()
@@ -399,20 +555,23 @@ def update_3_to_3_1(db: Connection) -> Connection:
             db_new.close()
         raise err
 
-    return connect_database("FA.db")
+    return connect_database(db_path)
 
 
 def update_3_1_to_3_2(db: Connection) -> Connection:
     print("Updating 3.1.0 to 3.2.0")
     db_new: Optional[Connection] = None
 
+    db_path: str = dp if (dp := database_path(db)) else "FA.db"
+    db_new_path: str = path_join(dirname(db_path), "new_" + basename(db_path))
+
     try:
-        db_new = connect_database("FA_new.db")
+        db_new = connect_database(db_new_path)
         make_database_3_2(db_new)
 
         # Transfer common submissions and users data
         print("Transfer common submissions and users data")
-        db.execute("ATTACH DATABASE 'FA_new.db' AS db_new")
+        db.execute(f"ATTACH DATABASE '{db_new_path}' AS db_new")
         db.execute(
             """INSERT OR IGNORE INTO db_new.SUBMISSIONS
             SELECT * FROM SUBMISSIONS"""
@@ -423,9 +582,8 @@ def update_3_1_to_3_2(db: Connection) -> Connection:
         )
         db.execute(
             """INSERT OR REPLACE INTO db_new.SETTINGS
-            SELECT * FROM SETTINGS"""
+            SELECT * FROM SETTINGS WHERE SETTING != "VERSION";"""
         )
-        db.execute("UPDATE db_new.SETTINGS SET SVALUE = '3.2.0' WHERE SETTING = 'VERSION'")
         db.execute("UPDATE db_new.USERS SET JOURNALS = ''")
 
         db.commit()
@@ -437,8 +595,8 @@ def update_3_1_to_3_2(db: Connection) -> Connection:
         db.close()
         db_new.commit()
         db_new.close()
-        move("FA.db", "FA_3_1.db")
-        move("FA_new.db", "FA.db")
+        move(db_path, path_join(dirname(db_path), "v3_1_" + basename(db_path)))
+        move(db_new_path, db_path)
     except (BaseException, Exception) as err:
         print("Database update interrupted!")
         db.commit()
@@ -448,54 +606,56 @@ def update_3_1_to_3_2(db: Connection) -> Connection:
             db_new.close()
         raise err
 
-    return connect_database("FA.db")
+    return connect_database(db_path)
 
 
 def update_3_2_to_3_3(db: Connection) -> Connection:
     print("Updating 3.2.0 to 3.3.0")
     db_new: Optional[Connection] = None
-    db_open: bool = True
+
+    db_path: str = dp if (dp := database_path(db)) else "FA.db"
+    db_new_path: str = path_join(dirname(db_path), "new_" + basename(db_path))
 
     try:
-        db_new = connect_database("FA_new.db")
-        make_tables(db_new)
+        db_new = connect_database(db_new_path)
+        make_database_3_3(db_new)
 
         # Transfer common submissions and users data
         print("Transfer common submissions and users data")
-        db.execute("ATTACH DATABASE 'FA_new.db' AS db_new")
+        db.execute(f"ATTACH DATABASE '{db_new_path}' AS db_new")
         db.execute(
-            f"""INSERT OR IGNORE INTO db_new.{users_table}
-            SELECT * FROM {users_table}"""
+            f"""INSERT OR IGNORE INTO db_new.USERS
+            SELECT * FROM USERS"""
         )
         db.execute(
-            f"""INSERT OR IGNORE INTO db_new.{submissions_table}
-            SELECT * FROM {submissions_table}"""
+            f"""INSERT OR IGNORE INTO db_new.SUBMISSIONS
+            SELECT * FROM SUBMISSIONS"""
         )
         db.execute(
-            f"""INSERT OR IGNORE INTO db_new.{journals_table}
-            SELECT * FROM {journals_table}"""
+            f"""INSERT OR IGNORE INTO db_new.JOURNALS
+            SELECT * FROM JOURNALS"""
         )
         db.execute(
-            f"""INSERT OR REPLACE INTO db_new.{settings_table}
-            SELECT * FROM {settings_table} WHERE SETTING != "LASTSTART" AND SETTING != "LASTUPDATE";"""
+            f"""INSERT OR REPLACE INTO db_new.SETTINGS
+            SELECT * FROM SETTINGS WHERE SETTING != "LASTSTART" AND SETTING != "LASTUPDATE" AND SETTING != "VERSION";"""
         )
-        db.execute("UPDATE db_new.SETTINGS SET SVALUE = '3.3.0' WHERE SETTING = 'VERSION'")
 
         # Add update to history
-        last_update: str = select(db, settings_table, ["SVALUE"], ["SETTING"], ["LASTUPDATE"]).fetchone()
+        last_update: str = db.execute("SELECT SVALUE FROM SETTINGS WHERE SETTING = 'LASTUPDATE'").fetchone()
         if last_update and last_update[0] != "0":
             # Commit and close databases to unlock
             db.commit()
             db.close()
             db = None
-            add_history(db_new, float(last_update[0]), "update")
+            db_new.execute("UPDATE SETTINGS SET SVALUE = ? WHERE SETTING = 'HISTORY'",
+                           json_dumps([[float(last_update[0]), "update"]]))
 
         # Close databases and replace old database
         print("Close databases and replace old database")
         db_new.commit()
         db_new.close()
-        move("FA.db", "FA_3_2.db")
-        move("FA_new.db", "FA.db")
+        move(db_path, path_join(dirname(db_path), "v3_2_" + basename(db_path)))
+        move(db_new_path, db_path)
     except (BaseException, Exception) as err:
         print("Database update interrupted!")
         if db is not None:
@@ -513,47 +673,49 @@ def update_3_4_to_3_5(db: Connection) -> Connection:
     print("Updating 3.4.0 to 3.5.0")
     db_new: Optional[Connection] = None
 
+    db_path: str = dp if (dp := database_path(db)) else "FA.db"
+    db_new_path: str = path_join(dirname(db_path), "new_" + basename(db_path))
+
     try:
-        db_new = connect_database("FA_new.db")
-        make_tables(db_new)
+        db_new = connect_database(db_new_path)
+        make_database_3_3(db_new)
 
         # Transfer common submissions and users data
         print("Transfer common submissions and users data")
-        db.execute("ATTACH DATABASE 'FA_new.db' AS db_new")
+        db.execute(f"ATTACH DATABASE '{db_new_path}' AS db_new")
         db.execute(
-            f"""INSERT OR IGNORE INTO db_new.{users_table}
-            SELECT * FROM {users_table}"""
+            f"""INSERT OR IGNORE INTO db_new.USERS
+            SELECT * FROM USERS"""
         )
         db.execute(
-            f"""INSERT OR IGNORE INTO db_new.{submissions_table}
-            SELECT * FROM {submissions_table}"""
+            f"""INSERT OR IGNORE INTO db_new.SUBMISSIONS
+            SELECT * FROM SUBMISSIONS"""
         )
         db.execute(
-            f"""INSERT OR IGNORE INTO db_new.{journals_table}
-            SELECT * FROM {journals_table}"""
+            f"""INSERT OR IGNORE INTO db_new.JOURNALS
+            SELECT * FROM JOURNALS"""
         )
         db.execute(
-            f"""INSERT OR REPLACE INTO db_new.{settings_table}
-            SELECT * FROM {settings_table} WHERE SETTING != "HISTORY";"""
+            f"""INSERT OR REPLACE INTO db_new.SETTINGS
+            SELECT * FROM SETTINGS WHERE SETTING != "HISTORY" AND SETTING != "VERSION";"""
         )
-        db.execute("UPDATE db_new.SETTINGS SET SVALUE = '3.5.0' WHERE SETTING = 'VERSION'")
 
         # Update history
         history: List[List[str]] = json_loads(
-            select(db, settings_table, ["SVALUE"], ["SETTING"], ["HISTORY"]).fetchone()[0]
+            db.execute("SELECT SVALUE FROM SETTINGS WHERE SETTING = 'HISTORY'").fetchone()[0]
         )
         history_new: List[Tuple[float, str]] = list(map(lambda th: (float(th[0]), th[1]), history))
         db.commit()
         db.close()
         db = None
-        write_setting(db_new, "HISTORY", json_dumps(history_new))
+        db_new.execute("UPDATE SETTINGS SET SVALUE = ? WHERE SETTING = 'HISTORY'", json_dumps(history_new))
 
         # Close databases and replace old database
         print("Close databases and replace old database")
         db_new.commit()
         db_new.close()
-        move("FA.db", "FA_3_4.db")
-        move("FA_new.db", "FA.db")
+        move(db_path, path_join(dirname(db_path), "v3_4_" + basename(db_path)))
+        move(db_new_path, db_path)
     except (BaseException, Exception) as err:
         print("Database update interrupted!")
         if db is not None:
@@ -564,12 +726,65 @@ def update_3_4_to_3_5(db: Connection) -> Connection:
             db_new.close()
         raise err
 
-    return connect_database("FA.db")
+    return connect_database(db_path)
+
+
+def update_3_8_to_4(db: Connection) -> Connection:
+    print("Updating 3.8.0 to 4.0.0")
+    db_new: Optional[Connection] = None
+
+    db_path: str = dp if (dp := database_path(db)) else "FA.db"
+    db_new_path: str = path_join(dirname(db_path), "new_" + basename(db_path))
+
+    try:
+        db_new = connect_database(db_new_path)
+        make_database_4(db_new)
+
+        # Transfer common submissions and users data
+        print("Transfer common submissions and users data")
+        db.execute(f"ATTACH DATABASE '{db_new_path}' AS db_new")
+        db.execute(
+            f"""INSERT OR IGNORE INTO db_new.USERS
+            SELECT * FROM USERS"""
+        )
+        db.execute(
+            f"""INSERT OR IGNORE INTO db_new.SUBMISSIONS
+            SELECT * FROM SUBMISSIONS"""
+        )
+        db.execute(
+            f"""INSERT OR IGNORE INTO db_new.JOURNALS
+            SELECT * FROM JOURNALS"""
+        )
+        db.execute(
+            f"""INSERT OR REPLACE INTO db_new.SETTINGS
+            SELECT * FROM SETTINGS WHERE SETTING != 'VERSION';"""
+        )
+
+        # Close databases and replace old database
+        print("Close databases and replace old database")
+        db.commit()
+        db.close()
+        db_new.commit()
+        db_new.close()
+        move(db_path, path_join(dirname(db_path), "v3_8_" + basename(db_path)))
+        move(db_new_path, db_path)
+    except (BaseException, Exception) as err:
+        print("Database update interrupted!")
+        if db is not None:
+            db.commit()
+            db.close()
+        if db_new is not None:
+            db_new.commit()
+            db_new.close()
+        raise err
+
+    return connect_database(db_path)
 
 
 def update_version(db: Connection, version: str, target_version: str) -> Connection:
     print(f"Updating {version} to {target_version}")
-    write_setting(db, "VERSION", target_version)
+    db.execute("UPDATE SETTINGS SET SVALUE = ? WHERE SETTING = 'VERSION'", [target_version])
+    db.commit()
     return db
 
 
@@ -594,11 +809,11 @@ def update_database(db: Connection) -> Connection:
         return update_database(update_version(db, db_version, "3.4.0"))  # 3.3.x to 3.4.0
     elif v >= 0 and (v := compare_versions(db_version, "3.5.0")) < 0:
         return update_database(update_3_4_to_3_5(db))  # 3.4.x to 3.5.0
-    elif v >= 0 and (v := compare_versions(db_version, "3.6.0")) < 0:
-        return update_database(update_version(db, db_version, "3.6.0"))  # 3.5.x to 3.6.0
-    elif v >= 0 and (v := compare_versions(db_version, "3.7.0")) < 0:
-        return update_database(update_version(db, db_version, "3.7.0"))  # 3.6.x to 3.7.0
+    elif v >= 0 and (v := compare_versions(db_version, "3.8.0")) < 0:
+        return update_database(update_version(db, db_version, "3.8.0"))  # 3.5.0-3.7.x to 3.8.0
+    elif v >= 0 and (v := compare_versions(db_version, "4.0.0")) < 0:
+        return update_database(update_3_8_to_4(db))  # 3.8.x to 4.0.0
     elif v >= 0 and compare_versions(db_version, __version__) < 0:
-        return update_version(db, db_version, __version__)
+        return update_version(db, db_version, __version__)  # Update to latest patch
 
     return db
